/*
 * sxpmroot.c : Sofken's xpmroot , put an Pixmap image on the root window .
 *
 * 1994/5/30 ver.0.01 created by yoneyama@(Business Brain Showa-ota Inc.)
 *              initial version.
 *
 * 1994/6/2  ver.0.02 modefied by ueda@white.tutkie.tut.ac.jp
 *              added file open error check.
 *              fixed bug about color.
 *
 * 1994/6/5  ver.0.03 modefied by tetu-s@is.aist-nara.ac.jp
 *              added randam select function.
 *              added X open display error check.
 *
 * 1994/6/5  ver.0.03a modefied by yoneyama@(Business Brain Showa-ota Inc.)
 *              added function coping with other Pixmaps.
 *              modefied the whole a little.
 * 1994/12/9 ver.0.04 modefied by tetu-s@is.aist-nara.ac.jp
 *              fixed bug about C-program comment.
 *              cope with XPM3 ( not perfect )
 * 1995/3/21 ver.0.05 modefied by tetu-s@is.aist-nara.ac.jp
 *              added functions to execute back ground.
 *              thanks to matuzaki@spica.c.chiba-u.ac.jp.
 *              debugging was supported by ueda@white.tutkie.tut.ac.jp .
 * 1995/12/11 ver.0.06 modefied by yoneyama@bbs.co.jp
 *              added option -f
 * 1996/6/15 ver.0.07 modefied by yoneyama@bbs.co.jp
 *              supported to read a compressed(gziped) XPM file.
 *
 * Copyleft by Software Research Academy.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>

static char *program_name = "sxpmroot";
static char *version      = "0.07";
static char *rvid = "$Id: sxpmroot.c ver.0.05 1995/3/21 Satoh Tetsu $";
static char *keytoken[] = {"c","m","s","g","g4",NULL};
static char *spacechr[] = {"a","\f","\n","\r","\t","\v",NULL};
struct XpmProperty *xp;

/* add a function for skip C-style comment to fgetc */
int Fgetc(f)
FILE *f;
{
    char c;
    c = fgetc(f);
    if(c == '/')
    {   c = fgetc(f);
        if(c == '*')
        { 
            do{ while(fgetc(f) != '*'); } while(fgetc(f)!='/');
            c = fgetc(f);
	}else
        {   ungetc(c,f);
            c = '/';
        }
    }
    return(c);
}

/* read a line from FILE *f,but if reading data is all space-like chara, *
 * try to read next line.                                                */
void Read_Data(b,s,f)
char *b;
int s;
FILE *f;
{
    int c,flag=0,sbuf;
    char *buf;
    buf = b;
    sbuf = s;
    while(isspace(c = Fgetc(f))); /* head space skip */
    while(--s && c!= EOF && c!= '\n')
    {
           *b++ = c;
           if(isgraph(c)) flag = 1;
           c = Fgetc(f);
    }
    if(c == '\n' && s) *b++='\n';
    *b = '\0';
    if(!flag) Read_Data(buf,sbuf,f);
}

/* if a program runs with illegal options , prints usages.*/
void usage()
{
    printf("%s version %s ", program_name, version);
    printf(" Copyleft by Software Research Academy.\n");
    printf("usage: %s [OPTIONS] file[s]\n", program_name);
    printf("  OPTIONS\n");
    printf("  -v\n");
    printf("  -display [{host}]:[{vs}]\n");
    printf("  -bg [second]\n");
    printf("  -f\n");
    exit(0);
}

/* convert X release 4 like hex formed strings to . */
int hex2rgb(p, red, green, blue)
char *p;
unsigned long int *red, *green, *blue;
{
    char *hex;
    int  len;
    
    len = (int)strlen(p + 1) / 3;
    hex = (char *)malloc(len + 1);

    strncpy(hex, p + 1, len);
    hex[len] = '\0';
    sscanf(hex, "%lx", red); 
    strncpy(hex, p + 1 + len, len);
    hex[len] = '\0';
    sscanf(hex, "%lx", green); 
    strncpy(hex, p + 1 + len * 2, len);
    hex[len] = '\0';
    sscanf(hex, "%lx", blue); 
    free(hex);

    *red   <<= (4 * (4 - len));
    *green <<= (4 * (4 - len));
    *blue  <<= (4 * (4 - len));

    return(0);
}

/* check whether string s start with a punctual string. */
int ispunc(s,punc)
char *s,**punc;
{
    int i=0,j,f;

    while(punc[i]!=NULL)
    {
        for(j=0,f=1;j<strlen(punc[i]);++j)
        {   if(*(s+j) != *(punc[i]+j))
            {   f = 0;
                break;
            }
        }
        if(f) return(1);
        i++;
    }
    return(0);
}

/* check whether string s is composed of digit characters. */
int isdigits(s)
char *s;
{
    int f = 0;
    while(*s != '\0')
    {   if(*s<'0' || *s>'9') f = 1;
        ++s;
    }
    return(f);
}
/* extract a string from s to d,as punc[] is a punctual string. */
char *extract(s,d,punc)
char *s,*d;
char **punc;
{
    char *e;
    e = s;
    do
    {   *d++ = *s++;
    }while(ispunc(s,punc));
    *d = '\0';
    
    while(*s!='\0') *e++ = *s++;
    *e = '\0';
    return(s);
}

void getKeyColor(buf,key,color)
char *buf,*key,*color;
{
    int j,k,l;
    char *s,*t,tokbuf[10];
    while(isspace(buf[0]))
    {    extract(buf,tokbuf,spacechr); /* tokbuf is dummy */
    }
    extract(buf,key,spacechr); /* extract key */
    while(isspace(buf[0]))
    {    extract(buf,tokbuf,spacechr); /* tokbuf is dummy */
    }
    for(j=0,s=NULL;keytoken[j]!=NULL && s==NULL;++j)
    {   
        for(k=0;k<strlen(buf)-2;++k)
        {
            if(isspace(buf[k]))
            {   l=1;
                for(;((k+l)<strlen(buf)-2)&&!isspace(buf[k+l]);++l)
                {   tokbuf[l-1] = buf[k+l];
                }
                tokbuf[l-1] = '\0';
                if(strcmp(tokbuf,keytoken[j])==0)
                {
                    s = &buf[k];
                    break;
                }
             }
        }
     }
    if(s==NULL)
    {   if((s=strstr(buf,"\""))==NULL)
        {   fprintf(stderr,"format error.\n");
            exit(1);
        }
    }
    for(t=buf;t<s;++t) *color++ = *t;
    strcpy(buf,s+1);
    do
    {
         *color--='\0';
    }while(isspace(*color));
}

struct XpmProperty
{
    char **xpmnames;
    int  numberOfXpmfiles;
    Bool verbose;
    char *dpyname;
    char *cc;
    XColor *xc;
    int width, height, ncols,perpixel;
    Display *d;
    Colormap cm;
    XColor dc;
    Pixmap pm;
    GC gc;
    Bool bg;
    int timeOfInterval;
    Bool force;
};

/* analyze arguments */
void analyzeOptions(argc,argv,xp)
int argc;
char **argv;
struct XpmProperty *xp;
{
    int i,j;
    if((xp->xpmnames = (char **)malloc(sizeof(char *) * argc))== NULL)
    {
        fprintf(stderr, "%s: cannot allocate memory.\n", program_name);
        exit(1);
    }
    for(i = 1,j = 0; i < argc; i++)
    {
        if(!strcmp(argv[i], "-display"))
        {
            xp->dpyname = argv[++i];
        }
        else if(!strcmp(argv[i],"-v"))
            xp->verbose = True;
        else if(!strcmp(argv[i],"-bg"))
        {    xp->bg = True;
             if(!isdigits(argv[i+1]))
             {
                 if((xp->timeOfInterval = atoi(argv[++i]))<0)
                 {    fprintf(stderr,
                     "%s: illegal value of interval\n",program_name);
                      exit(1);
                 }
             }else
	     {
                 xp->timeOfInterval = 10;
             }
        }
        else if(!strcmp(argv[i],"-f"))
        {
            xp->force = True;
        }else
        {
            xp->xpmnames[j++] = argv[i];
        }
    }

    if(j < 1)
    {
        usage();
    }
    xp->numberOfXpmfiles = j;
}

/* initialize the structure variable */
void initXpmProperty(xp)
struct XpmProperty *xp;
{
    xp->verbose = False;
    xp->dpyname = NULL;
    xp->bg = False;
    xp->force = False;
}
FILE *getXpmFile(xp)
struct XpmProperty *xp;
{
    FILE *fp; 
    char *xpmname;
    int len;
    xpmname = xp->xpmnames[rand() % xp->numberOfXpmfiles];
    if(xp->numberOfXpmfiles > 1 && xp->verbose) {
        printf("%s\n",xpmname);
    }

    len = strlen(xpmname);
    if(len > 0 && (xpmname[len - 1] == 'Z' || xpmname[len - 1] == 'z'))
    {
        char cmd[256];
        sprintf(cmd,"gzip -dc %s", xpmname);
        fp = popen(cmd, "r");
    }
    else
      fp = fopen(xpmname, "r");
    if(fp == NULL)
    {
        fprintf(stderr,"%s: cannot open pixmap file '%s'\n", 
                                      program_name, xpmname);
        usage();
    }
    return(fp);
}
void setImageProperty(buf,xp)
char *buf;
struct XpmProperty *xp;
{
    sscanf(&buf[1], "%d%d%d%d", 
             &xp->width, &xp->height, &xp->ncols,&xp->perpixel);
    /* get colors */
    xp->xc = (XColor *)malloc(sizeof(XColor)*xp->ncols);
    xp->cc = (char *)malloc(sizeof(char)*xp->ncols*(xp->perpixel+1));
}
void freeImageProperty(xp)
struct XpmProperty *xp;
{
    free(xp->xc);
    free(xp->cc);
}

/* allocate a color "colorp" to "i"th xc */
void allocColor(colorp,i,xp)
char *colorp;
int i;
struct XpmProperty *xp;
{
    unsigned long int red, green, blue;
    int flags = DoRed | DoGreen | DoBlue;
    if(colorp[0] == '#')
    {
	hex2rgb(colorp, &red, &green, &blue);
	xp->xc[i].red   = red;
	xp->xc[i].green = green;
	xp->xc[i].blue  = blue;
	if(xp->force)
	{   xp->xc[i].pixel = i;
            xp->xc[i].flags = flags;
	    XStoreColor(xp->d, xp->cm, &xp->xc[i]);
	}
	else if(!XAllocColor(xp->d, xp->cm, &xp->xc[i]))
        {   fprintf(stderr,"failed to allocate color (%s).\n",colorp);
        }
    }
    else if(strcmp(colorp,"None"))
    {
	if(xp->force)
	{
	    XStoreNamedColor(xp->d, xp->cm, colorp, i, &flags);
        }
	else if(!XAllocNamedColor(xp->d, xp->cm, colorp, &xp->xc[i], &xp->dc))
        {   fprintf(stderr,"failed to allocate color (%s).\n",colorp);
        }
    }
}
void drawOneLine(buf,i,xp)
char *buf;
int i;
struct XpmProperty *xp;
{
    int j,k;
    for(j = 0; j < xp->width; j++)
    {
        for(k = 0; k < xp->ncols; k++) 
        {    if(strncmp(xp->cc+k*(xp->perpixel+1),&buf[j*xp->perpixel + 1],xp->perpixel)==0) break;
        }
        XSetForeground(xp->d, xp->gc, xp->xc[k].pixel);
        XDrawPoint(xp->d, xp->pm, xp->gc, j, i);
    }
}
void setRootwindowXpmimage(sig)
int sig;
{
    FILE *fp;
    char *buf;
    Window r;
    Atom prop, type;
    int format;
    unsigned long length, after;
    unsigned char *data;
    int i, j, k;

    signal(sig,SIG_IGN);

    fp = getXpmFile(xp);

    if((xp->d = XOpenDisplay(xp->dpyname)) == NULL)
    {
    if((xp->d = XOpenDisplay(xp->dpyname)) == NULL)
    {    fprintf(stderr,"%s: failed opening display\n", program_name);
        if(xp->dpyname == NULL)
        {
            fprintf(stderr,"Please set environment variable DISPLAY \n");
            fprintf(stderr,"or specify display name using -display option.\n");
            fprintf(stderr,"Example 1:\n %% setenv DISPLAY penguin:0\n");
            fprintf(stderr,"Example 2:\n %% %s -display penguin:0 face1.xpm\n",
                                                      program_name);
         }
         exit(1);
     }
    }

    buf = (char *)malloc(256);
    Read_Data(buf, 256, fp); /* skip static char ... */
    Read_Data(buf, 256, fp); /* example:"128 128 8 1" */

    setImageProperty(buf,xp);
    r = RootWindow(xp->d, 0);
    if(xp->force)
        xp->cm = XCreateColormap(xp->d, r, DefaultVisual(xp->d, 0), AllocAll);
    else
        xp->cm = DefaultColormap(xp->d, 0);

    for(i = 0; i < xp->ncols; i++)
    {   char key[3],key1[3];
        char colorp[30],colorp1[30];
 
        Read_Data(buf,256,fp);
        
        strncpy(xp->cc+i*(xp->perpixel+1),&buf[1],xp->perpixel);
        *(xp->cc+i*(xp->perpixel+1)+xp->perpixel) = '\0';

        strcpy(buf,&buf[xp->perpixel+1]);

        getKeyColor(buf,key,colorp);
        if(strcmp(key,"c"))
        {   getKeyColor(buf,key1,colorp1);
            if(!strcmp(key1,"c"))
            {   strcpy(key,key1);
                strcpy(colorp,colorp1);
            }
        }
            

        if(!strcmp(key,"c") || !strcmp(key,"m"))
        {
            allocColor(colorp,i,xp);
        }
    }
    xp->pm = XCreatePixmap(xp->d, r, xp->width, xp->height, DefaultDepth(xp->d, 0));
    xp->gc = XCreateGC(xp->d, r, 0, 0);

    /* get a pixmap */
    buf = (char *)realloc(buf, xp->width*xp->perpixel + 6); /* width+'"'*2+','+'^M'+'\n'+'\0' */ 
    for(i = 0; i < xp->height; i++)
    {
        Read_Data(buf, xp->width*xp->perpixel + 6, fp);
        drawOneLine(buf,i,xp);
    }

    if(xp->force)
    {
        r = XCreateSimpleWindow(xp->d, RootWindow(xp->d, 0), 0, 0,
              DisplayWidth(xp->d,0), DisplayHeight(xp->d,0), 0, 0, 0);
        XMapWindow(xp->d, r);
        XSetWindowColormap(xp->d, r, xp->cm);
        XStoreName(xp->d, r, program_name);
    }
    XSetWindowBackgroundPixmap(xp->d, r, xp->pm);
    XClearWindow(xp->d, r);
    XFlush(xp->d);

    if(xp->force)
    {
        XSelectInput(xp->d, r, ButtonPressMask);
        while(1)
        {
            XEvent e;
            XNextEvent(xp->d, &e);
            if(e.type = ButtonPress) break;
        }
    }
    else
    {
        prop = XInternAtom(xp->d, "_XSETROOT_ID", False);
        XGetWindowProperty(xp->d, r, prop, 0L, 1L, True, AnyPropertyType,
            &type, &format, &length, &after, &data);
        if((type==XA_PIXMAP) && (format==32) && (length==1) && (after==0))
            XKillClient(xp->d, *((Pixmap *)data));
        XChangeProperty(xp->d, r, prop, XA_PIXMAP, 32, PropModeReplace,
            (unsigned char *) &xp->pm, 1);
        XSetCloseDownMode(xp->d, RetainPermanent);

    }
    XCloseDisplay(xp->d);
    freeImageProperty(xp);
    free(buf);
    fclose(fp);
    alarm(xp->timeOfInterval);
    signal(SIGALRM,setRootwindowXpmimage);
}


void main(argc, argv)
int argc;
char **argv;
{
    int pid;

    xp = (struct XpmProperty*)malloc(sizeof(struct XpmProperty));
    srand((int)time((time_t *)NULL));   
    initXpmProperty(xp);    
    analyzeOptions(argc,argv,xp);
    setRootwindowXpmimage(SIGALRM);
    if(xp->bg == True)
    {   pid = fork();
        if(pid == 0)
        {
            alarm(xp->timeOfInterval);
            signal(SIGALRM,setRootwindowXpmimage);
            pause();
        }
    }
}

